from graphene import ObjectType, String, Int
from ..utils import get_host, get_client_info


class GremlinClientInfo(ObjectType):
    gremlin_host = String()
    host_name = String()
    ip_address = String()

    def resolve_gremlin_host(self, info):
        return get_host(info.context['request'].app.state.gremlin_client.gremlin_server_url)

    def resolve_host_name(self, info):
        return get_client_info()['host_name']

    def resolve_ip_address(self, info):
        return get_client_info()['ip_address']


class LabelStats(ObjectType):
    label = String()
    count = Int()




import pytest
from .sample_payloads.vertex import CREATE_VERTICES_SAMPLES
from invana_engine.gremlin.core.element import VertexElement
from invana_engine.gremlin.core.exceptions import InvalidQueryArguments
import os

TEST_GRAPH_HOST = os.environ.get("TESTING_HOST")


class TestVertices:

    @pytest.fixture
    def graph_client(self):
        from invana_engine.gremlin import InvanaEngineClient
        return InvanaEngineClient(f"{TEST_GRAPH_HOST}/gremlin")

    def test_create_vertex(self, graph_client):
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            response = graph_client.vertex.create(label=vertex_sample["label"], properties=vertex_sample["properties"])
            assert isinstance(response, VertexElement)

    def test_read_vertex(self, graph_client):
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            props = vertex_sample['properties']
            responses = graph_client.vertex.read_many(
                label=vertex_sample['label'],
                query={list(props.keys())[0]: list(props.values())[0]})
            assert isinstance(responses, list)
            assert isinstance(responses[0], VertexElement)

    def test_update_vertex(self, graph_client):
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            props = vertex_sample['properties']
            response = graph_client.vertex.read_many(label=vertex_sample['label'],
                                                     query={
                                                         list(props.keys())[0]: list(props.values())[0]
                                                     })
            assert isinstance(response, list)
            response2 = graph_client.vertex.update(response[0].id, properties={"new_field": "yeah!"})
            assert isinstance(response2, VertexElement)

    def test_delete_single_vertex(self, graph_client):
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            props = vertex_sample['properties']
            responses = graph_client.vertex.read_many(
                label=vertex_sample['label'],
                query={
                    list(props.keys())[0]: list(props.values())[0]
                }
            )
            assert isinstance(responses, list)
            response2 = graph_client.vertex.delete_one(responses[0].id)

            responses3 = graph_client.vertex.read_one(responses[0].id)
            assert responses3 is None

    def test_delete_vertices(self, graph_client):
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            graph_client.vertex.create(label=vertex_sample['label'],
                                       properties=vertex_sample['properties'])

        graph_client.vertex.delete_many(label=list(CREATE_VERTICES_SAMPLES.keys())[0])
        responses2 = graph_client.vertex.read_many(label=list(CREATE_VERTICES_SAMPLES.keys())[0])
        assert responses2.__len__() == 0

    def test_delete_many_noargs_sent_error(self, graph_client):
        with pytest.raises(InvalidQueryArguments, match=r"label and query"):
            responses = graph_client.vertex.delete_many()

    def test_delete_one_noargs_sent_error(self, graph_client):
        with pytest.raises(TypeError, match=r"missing 1 required positional"):
            responses = graph_client.vertex.delete_one()

    def test_read_one_noargs_sent_error(self, graph_client):
        with pytest.raises(TypeError, match=r"missing 1 required positional"):
            responses = graph_client.vertex.read_one()



import graphene
from ..types.gremlin import GremlinClientInfo
from ..utils import get_host, get_client_info


class GenericClientInfoSchema:
    hello = graphene.String(name=graphene.String(default_value="World"))
    get_client_info = graphene.Field(GremlinClientInfo)

    def resolve_hello(self, info: graphene.ResolveInfo, name: any) -> str:
        return "Hello " + name

    def resolve_get_client_info(self, info: graphene.ResolveInfo):
        result = get_client_info()
        result['gremlin_host'] = get_host(info.context['request'].app.state.gremlin_client.gremlin_server_url)
        return result



"""


"""


class CreateElement:
    pass


class SerializedElement:
    id = None
    label = None
    type = None
    properties = {}

    @staticmethod
    def serialized_data(data, serializer=None):
        return serializer.serialize_data(data)

    def __init__(self, data, serializer=None):
        self.data = data
        serialized_data = self.serialized_data(data, serializer=serializer)
        self.assign_data(serialized_data)

    def assign_data(self, serialized_data):
        raise NotImplementedError()

    def assign_id_label_properties(self, id, label, properties):
        self.id = id
        self.label = label
        self.properties = properties

    def __dict__(self):
        return {
            "id": self.id,
            "label": self.label,
            "type": self.type,
            "properties": self.properties
        }

    def __repr__(self):
        return "<{type} label={label} id={id}> ".format(type=self.type.capitalize(),
                                                        id=self.id, label=self.label)

    def get_property(self, key):
        return self.properties.get(key)


class VertexElement(SerializedElement):
    type = "g:Vertex"

    def assign_data(self, serialized_data):
        self.assign_id_label_properties(serialized_data['id'], serialized_data['label'],
                                        serialized_data.get('properties', {}))


class EdgeElement(SerializedElement):
    type = "g:Edge"
    in_v = None
    in_v_label = None
    out_v = None
    out_v_label = None

    def assign_data(self, serialized_data):
        self.assign_id_label_properties(serialized_data['id'], serialized_data['label'],
                                        serialized_data.get('properties', {}),
                                        )
        self.in_v = serialized_data['inV']
        self.in_v_label = serialized_data['inVLabel']
        self.out_v = serialized_data['outV']
        self.out_v_label = serialized_data['outVLabel']

    def __dict__(self):
        return {
            "id": self.id,
            "label": self.label,
            "type": self.type,
            "properties": self.properties,
            "in_v": self.in_v,
            "out_v": self.out_v,
            'in_v_label': self.in_v_label,
            'out_v_label': self.out_v_label
        }



from graphene import Scalar, ObjectType, String, ID, DateTime, List, Boolean
import datetime


class AnyField(Scalar):

    @staticmethod
    def serialize(dt):
        if isinstance(dt, dict):
            transformed_data = {}
            for key, v in dt.items():
                if isinstance(v, datetime.datetime):
                    transformed_data[key] = v.isoformat()
                else:
                    transformed_data[key] = v
            return transformed_data
        elif isinstance(dt, datetime.datetime):
            return dt.isoformat()
        else:
            return dt

    @staticmethod
    def parse_literal(node):
        return node.value

    @staticmethod
    def parse_value(value):
        return value


class GrapheneVertexType(ObjectType):
    id = AnyField()
    type = String()
    label = String()
    properties = AnyField()


class GrapheneEdgeType(GrapheneVertexType):
    in_v = AnyField()
    out_v = AnyField()
    in_v_label = String()
    out_v_label = String()


class GrapheneVertexOrEdgeType(GrapheneEdgeType):
    pass


class ElementSchemaType(ObjectType):
    label = String()
    propertyKeys = AnyField()


class StatusType(ObjectType):
    status = Boolean()
    message = String()



from graphene import ObjectType, String, Field, JSONString, ResolveInfo, Int, List
from ..types.element import GrapheneVertexType, GrapheneEdgeType, AnyField, \
    GrapheneVertexOrEdgeType, ElementSchemaType, StatusType
from ..types.gremlin import LabelStats
from .client import GenericClientInfoSchema
from typing import Any

default_pagination_size = 10


class StatsManagement:
    get_vertices_label_stats = Field(List(LabelStats), namespace=String())
    get_edges_label_stats = Field(List(LabelStats), namespace=String())
    get_vertex_label_stats = Field(LabelStats, label=String())
    get_edge_label_stats = Field(LabelStats, label=String())

    def resolve_get_vertices_label_stats(self, info: ResolveInfo, namespace: str = None):
        data = info.context['request'].app.state.gremlin_client.stats.get_vertices_label_stats(namespace=namespace)
        return data

    def resolve_get_edges_label_stats(self, info: ResolveInfo, namespace: str = None):
        data = info.context['request'].app.state.gremlin_client.stats.get_edges_label_stats(namespace=namespace)
        return data

    def resolve_get_vertex_label_stats(self, info: ResolveInfo, label: str = None) -> any:
        return info.context['request'].app.state.gremlin_client.stats.get_vertex_label_stats(label)

    def resolve_get_edge_label_stats(self, info: ResolveInfo, label: str = None) -> any:
        return info.context['request'].app.state.gremlin_client.stats.get_edge_label_stats(label)


class SchemaManagement:
    get_vertex_label_schema = Field(ElementSchemaType, label=String())
    get_all_vertices_schema = Field(List(ElementSchemaType))
    get_edge_label_schema = Field(ElementSchemaType, label=String())
    get_all_edges_schema = Field(List(ElementSchemaType))

    create_vertex_label_schema = Field(StatusType, label=String())
    create_vertex_property_schema = Field(StatusType,
                                          label=String(),
                                          property_key=String(),
                                          data_type=String(),
                                          cardinality=String(),
                                          )
    create_edge_label_schema = Field(StatusType, label=String(), multiplicity=String())
    create_edge_property_schema = Field(StatusType,
                                        label=String(),
                                        property_key=String(),
                                        data_type=String(),
                                        cardinality=String(),
                                        )

    def resolve_get_vertex_label_schema(self, info: ResolveInfo, label: str = None) -> any:
        return info.context['request'].app.state.gremlin_client.schema.get_vertex_label_schema(label)

    def resolve_get_all_vertices_schema(self, info: ResolveInfo, ) -> any:
        return info.context['request'].app.state.gremlin_client.schema.get_all_vertices_schema()

    def resolve_get_edge_label_schema(self, info: ResolveInfo, label: str = None) -> any:
        return info.context['request'].app.state.gremlin_client.schema.get_edge_label_schema(label)

    def resolve_get_all_edges_schema(self, info: ResolveInfo, ) -> any:
        return info.context['request'].app.state.gremlin_client.schema.get_all_edges_schema()

    def resolve_create_vertex_label_schema(self, info: ResolveInfo, label: str) -> any:
        return info.context['request'].app.state.gremlin_client.schema.create_vertex_label_schema(label)

    def resolve_create_vertex_property_schema(self, info: ResolveInfo,
                                              label: str,
                                              property_key: str,
                                              data_type: str,
                                              cardinality: str = None,
                                              ) -> any:
        return info.context['request'].app.state.gremlin_client.schema.create_vertex_property_schema(
            label=label,
            property_key=property_key, data_type=data_type, cardinality=cardinality
        )

    def resolve_create_edge_label_schema(self, info: ResolveInfo, label: str, multiplicity: str = None) -> any:
        return info.context['request'].app.state.gremlin_client.schema.create_edge_label_schema(
            label, multiplicity=multiplicity)

    def resolve_create_edge_property_schema(self, info: ResolveInfo,
                                              label: str,
                                              property_key: str,
                                              data_type: str,
                                              cardinality: str = None,
                                              ) -> any:
        return info.context['request'].app.state.gremlin_client.schema.create_edge_property_schema(
            label=label,
            property_key=property_key,
            data_type=data_type,
            cardinality=cardinality
        )


class GremlinVertexQuerySchema:
    get_vertex_by_id = Field(GrapheneVertexType, id=String(required=True))
    filter_vertex = Field(List(GrapheneVertexType), label=String(), namespace=String(), query=JSONString(),
                          limit=Int(default_value=default_pagination_size), skip=Int())
    get_in_edges_and_vertices = Field(List(GrapheneVertexOrEdgeType), id=AnyField(required=True),
                                      label=String(), namespace=String(), query=JSONString(),
                                      limit=Int(default_value=default_pagination_size), skip=Int())
    get_out_edges_and_vertices = Field(List(GrapheneVertexOrEdgeType), id=AnyField(required=True),
                                       label=String(), namespace=String(), query=JSONString(),
                                       limit=Int(default_value=default_pagination_size), skip=Int())
    filter_vertex_and_neighbor_edges_and_vertices = Field(List(GrapheneVertexOrEdgeType), id=AnyField(),
                                            label=String(), namespace=String(), query=JSONString(),
                                            limit=Int(default_value=default_pagination_size), skip=Int())
    get_or_create_vertex = Field(GrapheneVertexType, label=String(), namespace=String(), properties=JSONString())

    def resolve_get_vertex_by_id(self, info: ResolveInfo, id: str):
        data = info.context['request'].app.state.gremlin_client.vertex.read_one(id)
        return data.__dict__() if data else None

    def resolve_filter_vertex(self, info: ResolveInfo, label: str = None, namespace: str = None, query: str = None,
                              limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.vertex.read_many(
            label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )
        return [datum.__dict__() for datum in data]

    def resolve_get_in_edges_and_vertices(self, info: ResolveInfo,
                                          id: Any = None, label: str = None, namespace: str = None,
                                          query: str = None,
                                          limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.vertex.read_in_edges_and_vertices(
            id, label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )
        return [datum.__dict__() for datum in data]

    def resolve_get_out_edges_and_vertices(self, info: ResolveInfo,
                                           id: Any = None, label: str = None, namespace: str = None, query: str = None,
                                           limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.vertex.read_out_edges_and_vertices(
            id, label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )
        return [datum.__dict__() for datum in data]

    def resolve_filter_vertex_and_neighbor_edges_and_vertices(self, info: ResolveInfo,
                                                id: Any = None, label: str = None, namespace: str = None,
                                                query: str = None,
                                                limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.vertex.filter_vertex_and_neighbor_edges_and_vertices(
            vertex_id=id, label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )
        return [datum.__dict__() for datum in data]

    def resolve_get_or_create_vertex(self, info: ResolveInfo, label: str = None, namespace: str = None,
                                     properties: str = None):
        return info.context['request'].app.state.gremlin_client.vertex.get_or_create(label=label,
                                                                                     namespace=namespace,
                                                                                     properties=properties)


class GremlinEdgeQuerySchema:
    get_edge_by_id = Field(GrapheneEdgeType, id=String(required=True))
    filter_edge = Field(List(GrapheneEdgeType), label=String(), query=JSONString(),
                        limit=Int(default_value=default_pagination_size), skip=Int())
    filter_edge_and_get_neighbor_vertices = Field(List(GrapheneEdgeType), label=String(), query=JSONString(),
                                                  limit=Int(default_value=default_pagination_size), skip=Int())
    get_or_create_edge = Field(GrapheneEdgeType, label=String(), namespace=String(), properties=JSONString())

    def resolve_get_edge_by_id(self, info: ResolveInfo, id: str):
        data = info.context['request'].app.state.gremlin_client.edge.read_one(id)
        return data.__dict__() if data else None

    def resolve_filter_edge(self, info: ResolveInfo, label: str = None, namespace: str = None,
                            query: str = None, limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.edge.read_many(
            label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )
        return [datum.__dict__() for datum in data]

    def resolve_filter_edge_and_get_neighbor_vertices(self, info: ResolveInfo, label: str = None,
                                                      namespace: str = None, query: str = None,
                                                      limit: int = default_pagination_size, skip: int = 0):
        data = info.context['request'].app.state.gremlin_client.edge.filter_edge_and_get_neighbor_vertices(
            label=label, namespace=namespace, query=query, limit=limit, skip=skip
        )

        #
        return [datum.__dict__() for datum in data]

    def resolve_get_or_create_edge(self, info: ResolveInfo, label: str = None, namespace: str = None,
                                   properties: str = None):
        return info.context['request'].app.state.gremlin_client.edge.get_or_create(label=label,
                                                                                   namespace=namespace,
                                                                                   properties=properties)


class GremlinRawQuerySchema:
    raw_query = Field(List(GrapheneEdgeType), gremlin=String())

    def resolve_raw_query(self, info: ResolveInfo, gremlin: str) -> any:
        return info.context['request'].app.state.gremlin_client.execute_query(gremlin)


class GremlinQuery(ObjectType,
                   StatsManagement,
                   SchemaManagement,
                   GremlinRawQuerySchema,
                   GremlinEdgeQuerySchema,
                   GremlinVertexQuerySchema,
                   GenericClientInfoSchema,
                   # GremlinManagementQuerySchema
                   ):
    pass



from gremlin_python.structure.graph import Vertex, Edge


class GraphSONV3Reader:

    @staticmethod
    def get_element_id(_id):
        if type(_id) is dict:
            _id = _id["@value"]
        if type(_id) is dict:
            _id = _id['relationId']
        if type(_id) is str:
            _id = _id.strip("#")
        return _id

    def serialize_element_dict(self, elem):
        cleaned_data = {"properties": {}}
        if "Direction.IN" in elem.keys():
            cleaned_data['type'] = "g:Edge"
        else:
            cleaned_data['type'] = "g:Vertex"
        for k, v in elem.items():
            if str(k) == "T.id":
                cleaned_data['id'] = self.get_element_id(v)
            elif str(k) == "T.label":
                cleaned_data['label'] = v
            elif str(k) == "Direction.OUT":
                _ = self.serialize_element_dict(v)
                cleaned_data.update({"outV": _['id'], "outVLabel": _['label']})
            elif str(k) == "Direction.IN":
                _ = self.serialize_element_dict(v)
                cleaned_data.update({"inV": _['id'], "inVLabel": _['label']})
            else:
                # TODO - check if this is right.
                cleaned_data['properties'][k] = v[0] if type(v) is list else v
        if cleaned_data['properties'].keys().__len__() == 0:
            del cleaned_data['properties']
        return cleaned_data

    @staticmethod
    def serialize_vertex_element(vertex):
        return {
            "id": vertex.id,
            "label": vertex.label,
            "type": "g:Vertex",
            "properties": {}
        }

    def serialize_edge_element(self, edge):
        return {
            "id": self.get_element_id(edge.id),
            "label": edge.label,
            "type": "g:Edge",
            "in_v": edge.inV.id,
            "in_v_label": edge.inV.label,
            "out_v": edge.outV.id,
            "out_v_label": edge.outV.label,
            "properties": {}
        }

    def serialize_data(self, data):
        if isinstance(data, list):
            _serialized_data = []
            for datum in data:
                _ = self.serialize_data(datum)
                if type(_) is list:
                    _serialized_data.extend(_)
                elif type(_) is dict:
                    _serialized_data.append(_)
            return _serialized_data
        elif isinstance(data, dict):
            return self.serialize_element_dict(data)
        elif isinstance(data, Vertex):
            return self.serialize_vertex_element(data)
        elif isinstance(data, Edge):
            return self.serialize_edge_element(data)
        else:
            raise NotImplementedError("")



import pytest
from .sample_payloads.vertex import CREATE_VERTICES_SAMPLES
from .sample_payloads.edge import CREATE_EDGES_SAMPLES
from invana_engine.gremlin.core.element import EdgeElement
from invana_engine.gremlin.core.exceptions import InvalidQueryArguments
import os

TEST_GRAPH_HOST = os.environ.get("TESTING_HOST")


class TestEdges:

    @pytest.fixture
    def graph_client(self):
        from invana_engine.gremlin import InvanaEngineClient
        return InvanaEngineClient(f"{TEST_GRAPH_HOST}/gremlin")

    @pytest.fixture
    def init_data(self):
        from invana_engine.gremlin import InvanaEngineClient
        graph_client = InvanaEngineClient(f"{TEST_GRAPH_HOST}/gremlin")
        for k, vertex_sample in CREATE_VERTICES_SAMPLES.items():
            graph_client.vertex.create(label=vertex_sample["label"], properties=vertex_sample["properties"])
        return graph_client.vertex.read_many()

    def test_create_edge(self, graph_client, init_data):
        for k, edge_sample in CREATE_EDGES_SAMPLES.items():
            graph_client.edge.create(
                label=edge_sample["label"],
                properties=edge_sample["properties"],
                inv=init_data[0].id,
                outv=init_data[1].id
            )

    def test_read_edge(self, graph_client):
        for k, edge_sample in CREATE_EDGES_SAMPLES.items():
            props = edge_sample['properties']
            responses = graph_client.edge.read_many(
                label=edge_sample['label'],
                query={list(props.keys())[0]: list(props.values())[0]})
            assert isinstance(responses, list)
            assert isinstance(responses[0], EdgeElement)

    def test_update_edge(self, graph_client):
        for k, edge_sample in CREATE_EDGES_SAMPLES.items():
            props = edge_sample['properties']
            response = graph_client.edge.read_many(
                label=edge_sample['label'],
                query={list(props.keys())[0]: list(props.values())[0]}
            )
            assert isinstance(response, list)
            response2 = graph_client.edge.update(response[0].id, properties={"new_field": "yeah!"})
            assert isinstance(response2, EdgeElement)

    def test_delete_single_edge(self, graph_client):
        for k, edge_sample in CREATE_EDGES_SAMPLES.items():
            props = edge_sample['properties']
            responses = graph_client.edge.read_many(
                label=edge_sample['label'],
                query={
                    list(props.keys())[0]: list(props.values())[0]
                }
            )
            assert isinstance(responses, list)
            response2 = graph_client.edge.delete_one(responses[0].id)
            responses3 = graph_client.edge.read_one(responses[0].id)
            assert responses3 is None

    def test_delete_edges(self, graph_client, init_data):
        for k, edge_sample in CREATE_EDGES_SAMPLES.items():
            graph_client.edge.create(
                label=edge_sample["label"],
                properties=edge_sample["properties"],
                inv=init_data[0].id,
                outv=init_data[1].id
            )
        responses = graph_client.edge.delete_many(label="has_satellite")
        responses2 = graph_client.edge.read_many(label="has_satellite")
        assert responses2.__len__() == 0

    def test_delete_many_noargs_sent_error(self, graph_client):
        with pytest.raises(InvalidQueryArguments, match=r"label and query"):
            responses = graph_client.edge.delete_many()

    def test_delete_one_noargs_sent_error(self, graph_client):
        with pytest.raises(TypeError, match=r"missing 1 required positional"):
            responses = graph_client.edge.delete_one()

    def test_read_one_noargs_sent_error(self, graph_client):
        with pytest.raises(TypeError, match=r"missing 1 required positional"):
            responses = graph_client.edge.read_one()



import abc
from .core.exceptions import InvalidQueryArguments


class GremlinOperationBase:

    def __init__(self, gremlin_client=None):
        """

        :param gremlin_client: InvanaEngineClient instance
        """
        self.gremlin_client = gremlin_client

    @staticmethod
    def get_namespaced_label(label=None, namespace=None):
        return "{}/{}".format(namespace, label) if namespace else label

    @property
    def serializer(self):
        return self.gremlin_client.serializer


class CRUDOperationsBase(GremlinOperationBase, metaclass=abc.ABCMeta):

    def filter_vertex(self, vertex_id=None, label=None, namespace=None, query=None, limit=None, skip=None):
        """

        :param vertex_id:
        :param label:
        :param namespace:
        :param query:
        :param limit:
        :param skip:
        :return:
        """
        label = self.get_namespaced_label(label=label, namespace=namespace)
        query = {} if query is None else query
        _ = self.gremlin_client.g.V(vertex_id) if vertex_id else self.gremlin_client.g.V()
        if label:
            _.hasLabel(label)
        if query:
            for k, v in query.items():
                _.has(k, v)
        if limit is not None and skip is not None:  # TODO - pagination fixes needed
            _.range(skip, skip + limit)
        return _

    def filter_edge(self, edge_id=None, label=None, namespace=None, query=None, limit=None, skip=None):
        """

        :param edge_id:
        :param label:
        :param namespace:
        :param query:
        :param limit:
        :param skip:
        :return:
        """
        label = self.get_namespaced_label(label=label, namespace=namespace)
        query = {} if query is None else query
        _ = self.gremlin_client.g.E(edge_id) if edge_id else self.gremlin_client.g.E()
        if label:
            _.hasLabel(label)
        for k, v in query.items():
            _.has(k, v)
        if limit is not None and skip is not None:  # TODO - pagination fixes needed
            _.range(skip, skip + limit)

        return _

    @abc.abstractmethod
    def create(self, label=None, namespace=None, properties=None, **kwargs):
        pass

    @abc.abstractmethod
    def update(self, elem_id, properties=None):
        pass

    @abc.abstractmethod
    def _read_one(self, element_id):
        pass

    def read_one(self, element_id):
        return self._read_one(element_id)

    @abc.abstractmethod
    def _read_many(self, label=None, namespace=None, query=None, limit=None, skip=None):
        pass

    def read_many(self, label=None, namespace=None, query=None, limit=None, skip=None):
        return self._read_many(label=label, namespace=namespace, query=query, limit=limit, skip=skip)

    @abc.abstractmethod
    def _delete_one(self, element_id):
        pass

    def delete_one(self, element_id):
        return self._delete_one(element_id)

    @staticmethod
    def validate_filter_many_filters(label=None, query=None):
        if label is None and query is None:
            raise InvalidQueryArguments("Both label and query arguments cannot be none")
        pass

    @abc.abstractmethod
    def _delete_many(self, label=None, namespace=None, query=None):
        pass

    def delete_many(self, label=None, namespace=None, query=None):
        self.validate_filter_many_filters(label=label, query=query)
        return self._delete_many(label=label, namespace=namespace, query=query)

    @abc.abstractmethod
    def get_or_create(self, label=None, namespace=None, properties=None):
        pass

    @staticmethod
    def drop(_):
        _.drop().iterate()



"""

"""

CREATE_EDGES_SAMPLES = {
    "has_satellite": {
        "label": "has_satellite",
        "properties": {
            "x_prop": "xyz",
        }
    },
    "has_animals": {
        "label": "has_animals",
        "properties": {
            "x_prop": "xyz",
        }
    }
}



from .edge import Edge
from .vertex import Vertex
from .stats import StatsOps
from .schema import SchemaOps
from .serializers.graphson_v3 import GraphSONV3Reader
from gremlin_python.process.anonymous_traversal import traversal
from gremlin_python.driver.driver_remote_connection import DriverRemoteConnection
from invana_engine.settings import gremlin_traversal_source as _gremlin_traversal_source


class GremlinClient:
    """

    client = GremlinClient("ws://127.0.0.1:8182/gremlin")
    client.execute_query("g.V().limit(1).toList()")
    """

    def __init__(self, gremlin_server_url,
                 gremlin_traversal_source=None,
                 gremlin_server_username=None,
                 gremlin_server_password=None,
                 serializer=None,
                 transport_factory=None):
        if gremlin_server_url is None:
            raise Exception("Invalid gremlin_server_url. default: ws://127.0.0.1:8182/gremlin")
        self.connection = DriverRemoteConnection(
            gremlin_server_url,
            gremlin_traversal_source or _gremlin_traversal_source,
            username=gremlin_server_username,
            password=gremlin_server_password,
            transport_factory=transport_factory
        )
        self.g = traversal().withRemote(self.connection)
        self.serializer = serializer

    def execute_query(self, raw_query, serialize_elements=True):
        """

        :param raw_query: Gremlin query in plain string.
        :param serialize_elements: this will convert data GraphSON data into JSON
        :return:
        """
        result = self.connection._client.submit(raw_query).all().result()
        if serialize_elements is True:
            return self.serializer.serialize_data(result)
        return result


class InvanaEngineClient:
    """
    Usage:

    graph_client = InvanaEngineClient(gremlin_server_url="ws://127.0.0.1:8182/gremlin")
    """

    def __init__(self,
                 gremlin_server_url,
                 gremlin_traversal_source=None,
                 gremlin_server_username=None,
                 gremlin_server_password=None,
                 serializer=None,
                 transport_factory=None):
        serializer = serializer or GraphSONV3Reader()
        self.gremlin_server_url = gremlin_server_url
        self.gremlin_client = GremlinClient(gremlin_server_url=gremlin_server_url,
                                            gremlin_traversal_source=gremlin_traversal_source,
                                            gremlin_server_username=gremlin_server_username,
                                            gremlin_server_password=gremlin_server_password,
                                            serializer=serializer,
                                            transport_factory=transport_factory)
        self.vertex = Vertex(gremlin_client=self.gremlin_client)
        self.edge = Edge(gremlin_client=self.gremlin_client)
        self.stats = StatsOps(gremlin_client=self.gremlin_client)
        self.schema = SchemaOps(gremlin_client=self.gremlin_client)

    def close(self):
        self.gremlin_client.connection.close()

    def delete_everything(self):
        """

        :return: None
        """
        return self.gremlin_client.g.V().drop().iterate()

    def execute_query(self, raw_query, serialize_elements=True):
        """

        :param raw_query: Gremlin query in plain string.
        :param serialize_elements: this will convert data GraphSON data into JSON
        :return:
        """
        return self.gremlin_client.execute_query(raw_query, serialize_elements=serialize_elements)



from gremlin_python.driver.protocol import GremlinServerWSProtocol, GremlinServerError
from gremlin_python.driver import request
from gremlin_python.driver.resultset import ResultSet

import base64


class InvanaGremlinServerWSProtocol(GremlinServerWSProtocol):

    def data_received(self, message, results_dict):
        # if Gremlin Server cuts off then we get a None for the message
        if message is None:
            raise GremlinServerError({
                'code': 500,
                'message': 'Server disconnected - please try to reconnect',
                'attributes': {}
            })

        message = self._message_serializer.deserialize_message(message)
        request_id = message['requestId']
        result_set = results_dict[request_id] if request_id in results_dict else ResultSet(None, None)
        status_code = message['status']['code']
        aggregate_to = message['result']['meta'].get('aggregateTo', 'list')
        data = message['result']['data']
        result_set.aggregate_to = aggregate_to
        if status_code == 407:
            auth = b''.join([b'\x00', self._username.encode('utf-8'),
                             b'\x00', self._password.encode('utf-8')])
            request_message = request.RequestMessage(
                'traversal', 'authentication',
                {'sasl': base64.b64encode(auth).decode()}
            )
            self.write(request_id, request_message)
            data = self._transport.read()
            # Allow recursive call for auth
            return self.data_received(data, results_dict)
        elif status_code == 204:
            result_set.stream.put_nowait([])
            del results_dict[request_id]
            return status_code
        elif status_code in [200, 206]:
            result_set.stream.put_nowait(data)
            if status_code == 200:
                result_set.status_attributes = message['status']['attributes']
                del results_dict[request_id]
            return status_code
        else:
            del results_dict[request_id]
            raise GremlinServerError(message["status"])
